import os
import sys
import inspect
from time import time
from copy import deepcopy

import numpy as np
from ase import Atoms
from ase.io import read, write
from flare.utils.learner import get_max_cutoff


class FLARE_Atoms(Atoms):
    """
    The `FLARE_Atoms` class is a child class of ASE `Atoms`, 
    which has completely the same usage as the primitive ASE `Atoms`, and
    in the meanwhile mimic `Structure` class. It is used in the `OTF` module
    with ASE engine (by `OTF_ASE` module). It enables attributes to be 
    obtained by both the name from ASE `Atoms` and `Structure`.

    The input arguments are the same as ASE `Atoms`.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prev_positions = np.zeros_like(self.positions)

    @staticmethod
    def from_ase_atoms(atoms):
        """
        Args:
            atoms (ASE Atoms): the ase atoms to build from
        """
        kw = [
            "symbols",  # use either "symbols" or "number"
            "positions",
            "tags",
            "momenta",
            "masses",
            "magmoms",
            "charges",
            "scaled_positions",
            "cell",
            "pbc",
            "celldisp",
            "constraint",
            "calculator",
            "info",
            "velocities",
        ]
        # The keywords are either the attr of atoms, or in dict atoms.arrays
        kwargs = {}
        for key in kw:
            try:
                kwargs[key] = getattr(atoms, key)
            except:
                if key in atoms.arrays:
                    kwargs[key] = atoms.arrays[key]
        kwargs["calculator"] = atoms.calc
        return FLARE_Atoms(**kwargs)

    @property
    def nat(self):
        return len(self)

    @property
    def species_labels(self):
        return self.symbols

    @property
    def coded_species(self):
        return self.numbers

    @property
    def forces(self):
        return self.get_forces()

    @forces.setter
    def forces(self, forces_array):
        pass

    @property
    def potential_energy(self):
        return self.get_potential_energy()

    @property
    def stress(self):
        return self.get_stress()

    @property
    def stress_stds(self):
        return None  # TODO: to implement

    @property
    def local_energy_stds(self):
        return None  # TODO: to implement

    @property
    def stds(self):
        try:  # when self.calc is not FLARE, there's no get_uncertainties()
            stds = self.calc.results["stds"]
        except:
            stds = np.zeros_like(self.positions)
        return stds

    def wrap_positions(self):
        return self.get_positions(wrap=True)

    @property
    def wrapped_positions(self):
        return self.get_positions(wrap=True)

    @property
    def max_cutoff(self):
        return get_max_cutoff(self.cell)

    def as_dict(self):
        return self.todict()

    @staticmethod
    def from_dict(dct):
        atoms = Atoms.fromdict(dct)
        return FLARE_Atoms.from_ase_atoms(atoms)



"""LAMMPS calculator for preparing and parsing single-point LAMMPS \
calculations."""
import subprocess
import numpy as np

# TODO: split LAMMPS input and data files into separate classes

def run_lammps(lammps_executable, input_file, output_file):
    """Runs a single point LAMMPS calculation.

    :param lammps_executable: LAMMPS executable file.
    :type lammps_executable: str
    :param input_file: LAMMPS input file.
    :type input_file: str
    :param output_file: Desired LAMMPS output file.
    :type output_file: str
    """
    # run lammps
    lammps_command = f'{lammps_executable} -in {input_file} '
    print("run command:", lammps_command)
    with open("tmp2False.out", "w+") as fout:
        subprocess.call(lammps_command.split(), stdout=fout)


def lammps_parser(dump_file, std=False):
    """Parses LAMMPS dump file. Assumes the forces are the final quantities \
to get dumped.

    :param dump_file: Dump file to be parsed.
    :type dump_file: str
    :return: Numpy array of forces on atoms.
    :rtype: np.ndarray
    """
    forces = []
    stds = []

    with open(dump_file, 'r') as outf:
        lines = outf.readlines()

    for count, line in enumerate(lines):
        if line.startswith('ITEM: ATOMS'):
            force_start = count

    for line in lines[force_start+1:]:
        fline = line.split()
        if std:
            forces.append([float(fline[-4]),
                           float(fline[-3]),
                           float(fline[-2])])
            stds.append(float(fline[-1]))
        else:
            forces.append([float(fline[-3]),
                           float(fline[-2]),
                           float(fline[-1])])

    return np.array(forces), np.array(stds)


# -----------------------------------------------------------------------------
#                           data functions
# -----------------------------------------------------------------------------

def lammps_dat(structure, atom_types, atom_masses, species):
    """Create LAMMPS data file for an uncharged material.

    :param structure: Structure object containing coordinates and cell.
    :type structure: struc.Structure
    :param atom_types: Atom types ranging from 1 to N.
    :type atom_types: List[int]
    :param atom_masses: Atomic masses of the atom types.
    :type atom_masses: List[int]
    :param species: Type of each atom.
    :type species: List[int]
    """

    dat_text = f"""Header of the LAMMPS data file

{structure.nat} atoms
{len(atom_types)} atom types
"""

    dat_text += lammps_cell_text(structure)
    dat_text += """
Masses

"""
    mass_text = ''
    for atom_type, atom_mass in zip(atom_types, atom_masses):
        mass_text += f'{atom_type} {atom_mass}\n'
    dat_text += mass_text
    dat_text += """
Atoms
"""
    dat_text += lammps_pos_text(structure, species)

    return dat_text


def lammps_dat_charged(structure, atom_types, atom_charges, atom_masses,
                       species):
    """Create LAMMPS data file for a charged material.

    :param structure: Structure object containing coordinates and cell.
    :type structure: struc.Structure
    :param atom_types: List of atom types.
    :type atom_types: List[int]
    :param atom_charges: Charge of each atom.
    :type atom_charges: List[float]
    :param atom_masses: Mass of each atom type.
    :type atom_masses: List[float]
    :param species: Type of each atom.
    :type species: List[int]
    """

    dat_text = f"""Header of the LAMMPS data file

{structure.nat} atoms
{len(atom_types)} atom types
"""

    dat_text += lammps_cell_text(structure)
    dat_text += """
Masses

"""
    mass_text = ''
    for atom_type, atom_mass in zip(atom_types,
                                    atom_masses):
        mass_text += f'{atom_type} {atom_mass}\n'
    dat_text += mass_text
    dat_text += """
Atoms
"""
    dat_text += lammps_pos_text_charged(structure, atom_charges, species)

    return dat_text


def lammps_cell_text(structure):
    """ Write cell from structure object."""

    cell_text = f"""
0.0 {structure.cell[0, 0]}  xlo xhi
0.0 {structure.cell[1, 1]}  ylo yhi
0.0 {structure.cell[2, 2]}  zlo zhi
{structure.cell[1, 0]} {structure.cell[2, 0]} {structure.cell[2, 1]}  xy xz yz
"""

    return cell_text


def lammps_pos_text(structure, species):
    """Create LAMMPS position text for a system of uncharged particles."""

    pos_text = '\n'
    for count, (pos, spec) in enumerate(zip(structure.positions, species)):
        pos_text += f'{count+1} {spec} {pos[0]} {pos[1]} {pos[2]}\n'
    return pos_text


def lammps_pos_text_charged(structure, charges, species):
    """Create LAMMPS position text for a system of charged particles."""

    pos_text = '\n'
    for count, (pos, chrg, spec) in enumerate(zip(structure.positions, charges,
                                                  species)):
        pos_text += f'{count+1} {spec} {chrg} {pos[0]} {pos[1]} {pos[2]}\n'
    return pos_text


def write_text(file, text):
    """Write text to file."""

    with open(file, 'w') as fin:
        fin.write(text)

# -----------------------------------------------------------------------------
#                           input functions
# -----------------------------------------------------------------------------


def generic_lammps_input(dat_file, style_string, coeff_string, dump_file, 
        newton=False, std_string=''):
    """Create text for generic LAMMPS input file."""

    if newton:
        ntn = 'on'
    else:
        ntn = 'off'

    if std_string != '':
        compute_cmd = f"compute std all uncertainty/atom {std_string}"
        c_std = "c_std"
    else:
        compute_cmd = ""
        c_std = ""

    input_text = f"""# generic lammps input file
units metal
atom_style atomic
dimension  3
boundary   p p p
newton {ntn}
read_data {dat_file}

pair_style {style_string}
pair_coeff {coeff_string}

thermo_style one
{compute_cmd}
dump 1 all custom 1 {dump_file} id type x y z fx fy fz {c_std}
dump_modify 1 sort id
run 0
"""

    return input_text


def ewald_input(dat_file, short_cut, kspace_accuracy, dump_file, newton=True):
    """Create text for Ewald input file."""
    if newton is True:
        ntn = 'on'
    else:
        ntn = 'off'

    input_text = f"""# Ewald input file
newton {ntn}
units metal
atom_style charge
dimension  3
boundary   p p p
read_data {dat_file}

pair_style coul/long {short_cut}
pair_coeff * *
kspace_style ewald {kspace_accuracy}

thermo_style one
dump 1 all custom 1 {dump_file} id type x y z fx fy fz
dump_modify 1 sort id
run 0
"""

    return input_text



"""
Implementation of three-body kernels using different cutoffs.

The kernels are slightly slower.
"""

import numpy as np
from numba import njit
from math import exp
import flare.kernels.cutoffs as cf
from flare.kernels.kernels import force_helper, grad_constants, grad_helper, \
    force_energy_helper, three_body_en_helper, three_body_helper_1, \
    three_body_helper_2, three_body_grad_helper_1, three_body_grad_helper_2

@njit
def three_body_mc_sepcut_jit(bond_array_1, c1, etypes1,
                             bond_array_2, c2, etypes2,
                             cross_bond_inds_1, cross_bond_inds_2,
                             cross_bond_dists_1, cross_bond_dists_2,
                             triplets_1, triplets_2,
                             d1, d2, sig, ls, r_cut, cutoff_func,
                             nspec, spec_mask, triplet_mask, cut3b_mask):

    kern = 0

    # pre-compute constants that appear in the inner loop
    sig2 = sig * sig
    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2

    bci = spec_mask[c1]
    bcin = bci * nspec
    bcj = spec_mask[c2]
    bcjn = bcj * nspec

    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        ci1 = bond_array_1[m, d1]
        ei1 = etypes1[m]

        # determine cutoff1 based on the end points
        bei1 = spec_mask[ei1]
        bei1n = nspec * bei1
        btype_ei1 = cut3b_mask[bei1 + bcin]
        cut_ei1 = r_cut[btype_ei1]
        fi1, fdi1 = cutoff_func(cut_ei1, ri1, ci1)

        for n in range(triplets_1[m]):

            ind1 = cross_bond_inds_1[m, m + n + 1]
            ri2 = bond_array_1[ind1, 0]
            ci2 = bond_array_1[ind1, d1]
            ei2 = etypes1[ind1]

            # skip if species does not match
            tr_spec = [c1, ei1, ei2]
            c2_ind = tr_spec
            if c2 in tr_spec:
                tr_spec.remove(c2)

                bei2 = spec_mask[ei2]
                btype_ei2 = cut3b_mask[bei2 + bcin]
                cut_ei2 = r_cut[btype_ei2]
                fi2, fdi2 = cutoff_func(cut_ei2, ri2, ci2)

                ttypei = triplet_mask[bcin + bei1n + bei2]
                tls1 = ls1[ttypei]
                tls2 = ls2[ttypei]
                tls3 = ls3[ttypei]
                tsig2 = sig2[ttypei]

                btype_ei3 = cut3b_mask[bei1n + bei2]
                cut_ei3 = r_cut[btype_ei3]
                ri3 = cross_bond_dists_1[m, m + n + 1]
                fi3, _ = cutoff_func(cut_ei3, ri3, 0)

                fi = fi1 * fi2 * fi3
                fdi = fdi1 * fi2 * fi3 + fi1 * fdi2 * fi3

                for p in range(bond_array_2.shape[0]):
                    ej1 = etypes2[p]

                    tr_spec1 = [tr_spec[0], tr_spec[1]]
                    if ej1 in tr_spec1:
                        tr_spec1.remove(ej1)

                        rj1 = bond_array_2[p, 0]
                        cj1 = bond_array_2[p, d2]
                        bej1 = spec_mask[ej1]
                        btype_ej1 = cut3b_mask[bej1 + bcjn]
                        cut_ej1 = r_cut[btype_ej1]
                        bej1n = nspec * bej1

                        fj1, fdj1 = cutoff_func(cut_ej1, rj1, cj1)

                        for q in range(triplets_2[p]):
                            ind2 = cross_bond_inds_2[p, p + 1 + q]
                            ej2 = etypes2[ind2]
                            if ej2 == tr_spec1[0]:
                                rj2 = bond_array_2[ind2, 0]
                                cj2 = bond_array_2[ind2, d2]

                                bej2 = spec_mask[ej2]
                                btype_ej2 = cut3b_mask[bej2 + bcjn]
                                cut_ej2 = r_cut[btype_ej2]

                                fj2, fdj2 = cutoff_func(cut_ej2, rj2, cj2)

                                btype_ej3 = cut3b_mask[bej1n + bej2]
                                cut_ej3 = r_cut[btype_ej3]
                                rj3 = cross_bond_dists_2[p, p + 1 + q]
                                fj3, _ = cutoff_func(cut_ej3, rj3, 0)

                                fj = fj1 * fj2 * fj3
                                fdj = fdj1 * fj2 * fj3 + fj1 * fdj2 * fj3

                                r11 = ri1 - rj1
                                r12 = ri1 - rj2
                                r13 = ri1 - rj3
                                r21 = ri2 - rj1
                                r22 = ri2 - rj2
                                r23 = ri2 - rj3
                                r31 = ri3 - rj1
                                r32 = ri3 - rj2
                                r33 = ri3 - rj3

                                if (c1 == c2):
                                    if (ei1 == ej1) and (ei2 == ej2):
                                        kern += \
                                            three_body_helper_1(ci1, ci2, cj1, cj2, r11,
                                                                r22, r33, fi, fj, fdi, fdj,
                                                                tls1, tls2, tls3,
                                                                tsig2)
                                    if (ei1 == ej2) and (ei2 == ej1):
                                        kern += \
                                            three_body_helper_1(ci1, ci2, cj2, cj1, r12,
                                                                r21, r33, fi, fj, fdi, fdj,
                                                                tls1, tls2, tls3,
                                                                tsig2)
                                if (c1 == ej1):
                                    if (ei1 == ej2) and (ei2 == c2):
                                        kern += \
                                            three_body_helper_2(ci2, ci1, cj2, cj1, r21,
                                                                r13, r32, fi, fj, fdi,
                                                                fdj,
                                                                tls1, tls2, tls3,
                                                                tsig2)
                                    if (ei1 == c2) and (ei2 == ej2):
                                        kern += \
                                            three_body_helper_2(ci1, ci2, cj2, cj1, r11,
                                                                r23, r32, fi, fj, fdi,
                                                                fdj,
                                                                tls1, tls2, tls3,
                                                                tsig2)
                                if (c1 == ej2):
                                    if (ei1 == ej1) and (ei2 == c2):
                                        kern += \
                                            three_body_helper_2(ci2, ci1, cj1, cj2, r22,
                                                                r13, r31, fi, fj, fdi,
                                                                fdj,
                                                                tls1, tls2, tls3,
                                                                tsig2)
                                    if (ei1 == c2) and (ei2 == ej1):
                                        kern += \
                                            three_body_helper_2(ci1, ci2, cj1, cj2, r12,
                                                                r23, r31, fi, fj, fdi,
                                                                fdj,
                                                                tls1, tls2, tls3,
                                                                tsig2)

    return kern


@njit
def three_body_mc_grad_sepcut_jit(bond_array_1, c1, etypes1,
                                  bond_array_2, c2, etypes2,
                                  cross_bond_inds_1, cross_bond_inds_2,
                                  cross_bond_dists_1, cross_bond_dists_2,
                                  triplets_1, triplets_2,
                                  d1, d2, sig, ls, r_cut, cutoff_func,
                                  nspec, spec_mask, ntriplet, triplet_mask, cut3b_mask):
    """Kernel gradient for 3-body force comparisons."""

    kern = 0.0
    sig_derv = np.zeros(ntriplet, dtype=np.float64)
    ls_derv = np.zeros(ntriplet, dtype=np.float64)

    # pre-compute constants that appear in the inner loop
    sig2, sig3, ls1, ls2, ls3, ls4, ls5, ls6 = grad_constants(sig, ls)

    bc1 = spec_mask[c1]
    bc2 = spec_mask[c2]
    bc1nn = bc1 * nspec * nspec

    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        ci1 = bond_array_1[m, d1]
        ei1 = etypes1[m]

        bei1 = spec_mask[ei1]
        bei1n = bei1 * nspec
        bond_m = cut3b_mask[bei1n+bc1]
        fi1, fdi1 = cutoff_func(r_cut[bond_m], ri1, ci1)

        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m + n + 1]
            ri3 = cross_bond_dists_1[m, m + n + 1]
            ri2 = bond_array_1[ind1, 0]
            ci2 = bond_array_1[ind1, d1]
            ei2 = etypes1[ind1]

            # skip if species does not match
            tr_spec = [c1, ei1, ei2]
            c2_ind = tr_spec
            if c2 in tr_spec:
                tr_spec.remove(c2)

                bei2 = spec_mask[ei2]

                ttypei = triplet_mask[bc1nn + bei1n + bei2]

                tls1 = ls1[ttypei]
                tls2 = ls2[ttypei]
                tls3 = ls3[ttypei]
                tls4 = ls4[ttypei]
                tls5 = ls5[ttypei]
                tls6 = ls6[ttypei]
                tsig2 = sig2[ttypei]
                tsig3 = sig3[ttypei]

                bond_n = cut3b_mask[bei2*nspec+bc1]
                fi2, fdi2 = cutoff_func(r_cut[bond_n], ri2, ci2)

                bond_mn = cut3b_mask[bei2+bei1n]
                fi3, _ = cutoff_func(r_cut[bond_mn], ri3, 0)

                fi = fi1 * fi2 * fi3
                fdi = fdi1 * fi2 * fi3 + fi1 * fdi2 * fi3

                for p in range(bond_array_2.shape[0]):
                    ej1 = etypes2[p]

                    tr_spec1 = [tr_spec[0], tr_spec[1]]
                    if ej1 in tr_spec1:
                        tr_spec1.remove(ej1)

                        rj1 = bond_array_2[p, 0]
                        cj1 = bond_array_2[p, d2]
                        bej1 = spec_mask[ej1]
                        bej1n = spec_mask[ej1]*nspec

                        bond_p = cut3b_mask[bc2+bej1n]

                        fj1, fdj1 = cutoff_func(r_cut[bond_p], rj1, cj1)

                        for q in range(triplets_2[p]):
                            ind2 = cross_bond_inds_2[p, p + q + 1]
                            ej2 = etypes2[ind2]
                            if ej2 == tr_spec1[0]:
                                rj3 = cross_bond_dists_2[p, p + q + 1]
                                rj2 = bond_array_2[ind2, 0]
                                cj2 = bond_array_2[ind2, d2]

                                bej2 = spec_mask[ej2]
                                bond_q = cut3b_mask[bc2+bej2*nspec]
                                fj2, fdj2 = cutoff_func(r_cut[bond_q], rj2, cj2)

                                bond_pq = cut3b_mask[bej1n + bej2]
                                fj3, _ = cutoff_func(r_cut[bond_pq], rj3, 0)

                                fj = fj1 * fj2 * fj3
                                fdj = fdj1 * fj2 * fj3 + fj1 * fdj2 * fj3

                                r11 = ri1 - rj1
                                r12 = ri1 - rj2
                                r13 = ri1 - rj3
                                r21 = ri2 - rj1
                                r22 = ri2 - rj2
                                r23 = ri2 - rj3
                                r31 = ri3 - rj1
                                r32 = ri3 - rj2
                                r33 = ri3 - rj3

                                if (c1 == c2):
                                    if (ei1 == ej1) and (ei2 == ej2):
                                        kern_term, sig_term, ls_term = \
                                            three_body_grad_helper_1(ci1, ci2, cj1, cj2,
                                                                     r11, r22, r33, fi, fj,
                                                                     fdi, fdj, tls1, tls2,
                                                                     tls3, tls4, tls5,
                                                                     tls6,
                                                                     tsig2, tsig3)
                                        kern += kern_term
                                        sig_derv[ttypei] += sig_term
                                        ls_derv[ttypei] += ls_term

                                    if (ei1 == ej2) and (ei2 == ej1):
                                        kern_term, sig_term, ls_term = \
                                            three_body_grad_helper_1(ci1, ci2, cj2, cj1,
                                                                     r12, r21, r33, fi, fj,
                                                                     fdi, fdj, tls1, tls2,
                                                                     tls3, tls4, tls5,
                                                                     tls6,
                                                                     tsig2, tsig3)
                                        kern += kern_term
                                        sig_derv[ttypei] += sig_term
                                        ls_derv[ttypei] += ls_term

                                if (c1 == ej1):
                                    if (ei1 == ej2) and (ei2 == c2):
                                        kern_term, sig_term, ls_term = \
                                            three_body_grad_helper_2(ci2, ci1, cj2, cj1,
                                                                     r21, r13, r32, fi, fj,
                                                                     fdi, fdj, tls1, tls2,
                                                                     tls3, tls4, tls5,
                                                                     tls6,
                                                                     tsig2, tsig3)
                                        kern += kern_term
                                        sig_derv[ttypei] += sig_term
                                        ls_derv[ttypei] += ls_term

                                    if (ei1 == c2) and (ei2 == ej2):
                                        kern_term, sig_term, ls_term = \
                                            three_body_grad_helper_2(ci1, ci2, cj2, cj1,
                                                                     r11, r23, r32, fi, fj,
                                                                     fdi, fdj, tls1, tls2,
                                                                     tls3, tls4, tls5,
                                                                     tls6,
                                                                     tsig2, tsig3)
                                        kern += kern_term
                                        sig_derv[ttypei] += sig_term
                                        ls_derv[ttypei] += ls_term

                                if (c1 == ej2):
                                    if (ei1 == ej1) and (ei2 == c2):
                                        kern_term, sig_term, ls_term = \
                                            three_body_grad_helper_2(ci2, ci1, cj1, cj2,
                                                                     r22, r13, r31, fi, fj,
                                                                     fdi, fdj, tls1, tls2,
                                                                     tls3, tls4, tls5,
                                                                     tls6,
                                                                     tsig2, tsig3)
                                        kern += kern_term
                                        sig_derv[ttypei] += sig_term
                                        ls_derv[ttypei] += ls_term

                                    if (ei1 == c2) and (ei2 == ej1):
                                        kern_term, sig_term, ls_term = \
                                            three_body_grad_helper_2(ci1, ci2, cj1, cj2,
                                                                     r12, r23, r31, fi, fj,
                                                                     fdi, fdj, tls1, tls2,
                                                                     tls3, tls4, tls5,
                                                                     tls6,
                                                                     tsig2, tsig3)

                                        kern += kern_term
                                        sig_derv[ttypei] += sig_term
                                        ls_derv[ttypei] += ls_term

    kern_grad = np.hstack((sig_derv, ls_derv))
    # np.zeros(2*ntriplet, dtype=np.float64)
    # kern_grad[:ntriplet] = sig_derv
    # kern_grad[ntriplet:] = ls_derv

    return kern, kern_grad


@njit
def three_body_mc_force_en_sepcut_jit(bond_array_1, c1, etypes1,
                                      bond_array_2, c2, etypes2,
                                      cross_bond_inds_1, cross_bond_inds_2,
                                      cross_bond_dists_1, cross_bond_dists_2,
                                      triplets_1, triplets_2,
                                      d1, sig, ls, r_cut, cutoff_func,
                                      nspec, spec_mask, triplet_mask, cut3b_mask):
    """Kernel for 3-body force/energy comparisons."""

    kern = 0

    # pre-compute constants that appear in the inner loop
    sig2 = sig * sig
    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)

    bc1 = spec_mask[c1]
    bc2 = spec_mask[c2]
    bc1nn = nspec * nspec * bc1

    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        ci1 = bond_array_1[m, d1]
        ei1 = etypes1[m]

        bei1 = spec_mask[ei1]
        bei1n = nspec * bei1

        bond_m = cut3b_mask[bc1+bei1n]

        fi1, fdi1 = cutoff_func(r_cut[bond_m], ri1, ci1)

        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m + n + 1]
            ei2 = etypes1[ind1]

            # skip if species does not match
            tr_spec = [c1, ei1, ei2]
            c2_ind = tr_spec
            if c2 in tr_spec:
                tr_spec.remove(c2)

                ri2 = bond_array_1[ind1, 0]
                ci2 = bond_array_1[ind1, d1]
                bei2 = spec_mask[ei2]

                ttypei = triplet_mask[bc1nn + bei1n + bei2]

                tls1 = ls1[ttypei]
                tls2 = ls2[ttypei]
                tsig2 = sig2[ttypei]

                bond_n = cut3b_mask[bc1*nspec+bei2]
                fi2, fdi2 = cutoff_func(r_cut[bond_n], ri2, ci2)

                ri3 = cross_bond_dists_1[m, m + n + 1]
                bond_mn = cut3b_mask[bei1n+bei2]
                fi3, _ = cutoff_func(r_cut[bond_mn], ri3, 0)

                fi = fi1 * fi2 * fi3
                fdi = fdi1 * fi2 * fi3 + fi1 * fdi2 * fi3

                for p in range(bond_array_2.shape[0]):

                    ej1 = etypes2[p]
                    tr_spec1 = [tr_spec[0], tr_spec[1]]
                    if ej1 in tr_spec1:
                        tr_spec1.remove(ej1)

                        rj1 = bond_array_2[p, 0]
                        bej1 = spec_mask[ej1]
                        bej1n = nspec*bej1

                        bond_p = cut3b_mask[bej1n+bc2]

                        fj1, _ = cutoff_func(r_cut[bond_p], rj1, 0)

                        tr_spec1 = [tr_spec[0], tr_spec[1]]
                        if ej1 in tr_spec1:
                            tr_spec1.remove(ej1)

                            for q in range(triplets_2[p]):
                                ind2 = cross_bond_inds_2[p, p + q + 1]
                                ej2 = etypes2[ind2]
                                if ej2 == tr_spec1[0]:
                                    rj2 = bond_array_2[ind2, 0]
                                    rj3 = cross_bond_dists_2[p, p + q + 1]

                                    bej2 = spec_mask[ej2]
                                    bond_q = cut3b_mask[bej2+bc2*nspec]
                                    fj2, _ = cutoff_func(r_cut[bond_q], rj2, 0)

                                    bond_pq = cut3b_mask[bej2+bej1n]
                                    fj3, _ = cutoff_func(r_cut[bond_pq], rj3, 0)

                                    fj = fj1 * fj2 * fj3

                                    r11 = ri1 - rj1
                                    r12 = ri1 - rj2
                                    r13 = ri1 - rj3
                                    r21 = ri2 - rj1
                                    r22 = ri2 - rj2
                                    r23 = ri2 - rj3
                                    r31 = ri3 - rj1
                                    r32 = ri3 - rj2
                                    r33 = ri3 - rj3

                                    if (c1 == c2):
                                        if (ei1 == ej1) and (ei2 == ej2):
                                            kern += three_body_en_helper(ci1, ci2, r11, r22,
                                                                         r33, fi, fj, fdi,
                                                                         tls1,
                                                                         tls2, tsig2)
                                        if (ei1 == ej2) and (ei2 == ej1):
                                            kern += three_body_en_helper(ci1, ci2, r12, r21,
                                                                         r33, fi, fj, fdi,
                                                                         tls1,
                                                                         tls2, tsig2)
                                    if (c1 == ej1):
                                        if (ei1 == ej2) and (ei2 == c2):
                                            kern += three_body_en_helper(ci1, ci2, r13, r21,
                                                                         r32, fi, fj, fdi,
                                                                         tls1,
                                                                         tls2, tsig2)
                                        if (ei1 == c2) and (ei2 == ej2):
                                            kern += three_body_en_helper(ci1, ci2, r11, r23,
                                                                         r32, fi, fj, fdi,
                                                                         tls1,
                                                                         tls2, tsig2)
                                    if (c1 == ej2):
                                        if (ei1 == ej1) and (ei2 == c2):
                                            kern += three_body_en_helper(ci1, ci2, r13, r22,
                                                                         r31, fi, fj, fdi,
                                                                         tls1,
                                                                         tls2, tsig2)
                                        if (ei1 == c2) and (ei2 == ej1):
                                            kern += three_body_en_helper(ci1, ci2, r12, r23,
                                                                         r31, fi, fj, fdi,
                                                                         tls1,
                                                                         tls2, tsig2)

    return kern


@njit
def three_body_mc_en_sepcut_jit(bond_array_1, c1, etypes1,
                                bond_array_2, c2, etypes2,
                                cross_bond_inds_1, cross_bond_inds_2,
                                cross_bond_dists_1, cross_bond_dists_2,
                                triplets_1, triplets_2,
                                sig, ls, r_cut, cutoff_func,
                                nspec, spec_mask, triplet_mask, cut3b_mask):
    kern = 0

    sig2 = sig * sig
    ls2 = 1 / (2 * ls * ls)

    bc1 = spec_mask[c1]
    bc1nn = bc1 * nspec * nspec
    bc2 = spec_mask[c2]

    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        ei1 = etypes1[m]

        bei1 = spec_mask[ei1]
        bei1n = nspec * bei1

        bond_m = cut3b_mask[bei1n + bc1]

        fi1, _ = cutoff_func(r_cut[bond_m], ri1, 0)

        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m + n + 1]
            ri2 = bond_array_1[ind1, 0]
            ei2 = etypes1[ind1]

            # skip if species does not match
            tr_spec = [c1, ei1, ei2]
            c2_ind = tr_spec
            if c2 in tr_spec:
                tr_spec.remove(c2)

                bei2 = spec_mask[ei2]

                ttypei = triplet_mask[bc1nn + bei1n + bei2]

                tls2 = ls2[ttypei]
                tsig2 = sig2[ttypei]

                ri3 = cross_bond_dists_1[m, m + n + 1]

                bond_n = cut3b_mask[bei2 + bc1*nspec]
                fi2, _ = cutoff_func(r_cut[bond_n], ri2, 0)

                bond_mn = cut3b_mask[bei1n + bei2]
                fi3, _ = cutoff_func(r_cut[bond_mn], ri3, 0)

                fi = fi1 * fi2 * fi3

                for p in range(bond_array_2.shape[0]):

                    ej1 = etypes2[p]
                    tr_spec1 = [tr_spec[0], tr_spec[1]]
                    if ej1 in tr_spec1:
                        tr_spec1.remove(ej1)

                        rj1 = bond_array_2[p, 0]
                        bej1n = spec_mask[ej1]*nspec
                        bond_p = cut3b_mask[bc2 + bej1n]
                        fj1, _ = cutoff_func(r_cut[bond_p], rj1, 0)

                        for q in range(triplets_2[p]):
                            ind2 = cross_bond_inds_2[p, p + q + 1]
                            ej2 = etypes2[ind2]
                            if ej2 == tr_spec1[0]:
                                rj2 = bond_array_2[ind2, 0]

                                bej2 = spec_mask[ej2]
                                bond_q = cut3b_mask[bc2+bej2*nspec]
                                fj2, _ = cutoff_func(r_cut[bond_q], rj2, 0)

                                rj3 = cross_bond_dists_2[p, p + q + 1]
                                bond_pq = cut3b_mask[bej1n+bej2]
                                fj3, _ = cutoff_func(r_cut[bond_pq], rj3, 0)
                                fj = fj1 * fj2 * fj3

                                r11 = ri1 - rj1
                                r12 = ri1 - rj2
                                r13 = ri1 - rj3
                                r21 = ri2 - rj1
                                r22 = ri2 - rj2
                                r23 = ri2 - rj3
                                r31 = ri3 - rj1
                                r32 = ri3 - rj2
                                r33 = ri3 - rj3

                                if (c1 == c2):
                                    if (ei1 == ej1) and (ei2 == ej2):
                                        C1 = r11 * r11 + r22 * r22 + r33 * r33
                                        kern += tsig2 * exp(-C1 * tls2) * fi * fj
                                    if (ei1 == ej2) and (ei2 == ej1):
                                        C3 = r12 * r12 + r21 * r21 + r33 * r33
                                        kern += tsig2 * exp(-C3 * tls2) * fi * fj
                                if (c1 == ej1):
                                    if (ei1 == ej2) and (ei2 == c2):
                                        C5 = r13 * r13 + r21 * r21 + r32 * r32
                                        kern += tsig2 * exp(-C5 * tls2) * fi * fj
                                    if (ei1 == c2) and (ei2 == ej2):
                                        C2 = r11 * r11 + r23 * r23 + r32 * r32
                                        kern += tsig2 * exp(-C2 * tls2) * fi * fj
                                if (c1 == ej2):
                                    if (ei1 == ej1) and (ei2 == c2):
                                        C6 = r13 * r13 + r22 * r22 + r31 * r31
                                        kern += tsig2 * exp(-C6 * tls2) * fi * fj
                                    if (ei1 == c2) and (ei2 == ej1):
                                        C4 = r12 * r12 + r23 * r23 + r31 * r31
                                        kern += tsig2 * exp(-C4 * tls2) * fi * fj

    return kern



import numpy as np
from numba import njit
from flare.env import AtomicEnvironment
from typing import Callable
import flare.cutoffs as cf
from math import exp


class ManyBodyKernel:
    def __init__(self, hyperparameters: 'ndarray', cutoff: float,
                 cutoff_func: Callable = cf.quadratic_cutoff):
        self.hyperparameters = hyperparameters
        self.signal_variance = hyperparameters[0]
        self.length_scale = hyperparameters[1]
        self.cutoff = cutoff
        self.cutoff_func = cutoff_func



"""
Utility functions for various tasks.
"""
from warnings import warn
from json import JSONEncoder
from typing import List
from math import inf

import numpy as np

_user_element_to_Z = {}
_user_Z_to_element = {}


def inject_user_definition(element: str, Z: int):
    """ Allow user-defined element. The definition
    will override the default ones from the periodic table.

    Example:

    >>> import flare.utils
    >>> import flare.utils.element_coder as ec
    >>> ec.inject_user_definition('C1', 6)
    >>> ec.inject_user_definition('C2', 7)
    >>> ec.inject_user_definition('H1', 1)
    >>> ec.inject_user_definition('H2', 2)

    This block should be executed before any other
    flare modules are imported. And user has to
    be very careful to not let Z overlap with other
    elements in the system

    :param element: string symbol of the element
    :type element: str
    :param Z: corresponding Z
    :type Z: int
    """
    _user_element_to_Z[element] = Z
    _user_Z_to_element[Z] = element


# Dictionary mapping elements to their atomic number (Z)
_element_to_Z = {'H': 1,
                 'He': 2,
                 'Li': 3,
                 'Be': 4,
                 'B': 5,
                 'C': 6,
                 'N': 7,
                 'O': 8,
                 'F': 9,
                 'Ne': 10,
                 'Na': 11,
                 'Mg': 12,
                 'Al': 13,
                 'Si': 14,
                 'P': 15,
                 'S': 16,
                 'Cl': 17,
                 'Ar': 18,
                 'K': 19,
                 'Ca': 20,
                 'Sc': 21,
                 'Ti': 22,
                 'V': 23,
                 'Cr': 24,
                 'Mn': 25,
                 'Fe': 26,
                 'Co': 27,
                 'Ni': 28,
                 'Cu': 29,
                 'Zn': 30,
                 'Ga': 31,
                 'Ge': 32,
                 'As': 33,
                 'Se': 34,
                 'Br': 35,
                 'Kr': 36,
                 'Rb': 37,
                 'Sr': 38,
                 'Y': 39,
                 'Zr': 40,
                 'Nb': 41,
                 'Mo': 42,
                 'Tc': 43,
                 'Ru': 44,
                 'Rh': 45,
                 'Pd': 46,
                 'Ag': 47,
                 'Cd': 48,
                 'In': 49,
                 'Sn': 50,
                 'Sb': 51,
                 'Te': 52,
                 'I': 53,
                 'Xe': 54,
                 'Cs': 55,
                 'Ba': 56,
                 'La': 57,
                 'Ce': 58,
                 'Pr': 59,
                 'Nd': 60,
                 'Pm': 61,
                 'Sm': 62,
                 'Eu': 63,
                 'Gd': 64,
                 'Tb': 65,
                 'Dy': 66,
                 'Ho': 67,
                 'Er': 68,
                 'Tm': 69,
                 'Yb': 70,
                 'Lu': 71,
                 'Hf': 72,
                 'Ta': 73,
                 'W': 74,
                 'Re': 75,
                 'Os': 76,
                 'Ir': 77,
                 'Pt': 78,
                 'Au': 79,
                 'Hg': 80,
                 'Tl': 81,
                 'Pb': 82,
                 'Bi': 83,
                 'Po': 84,
                 'At': 85,
                 'Rn': 86,
                 'Fr': 87,
                 'Ra': 88,
                 'Ac': 89,
                 'Th': 90,
                 'Pa': 91,
                 'U': 92,
                 'Np': 93,
                 'Pu': 94,
                 'Am': 95,
                 'Cm': 96,
                 'Bk': 97,
                 'Cf': 98,
                 'Es': 99,
                 'Fm': 100,
                 'Md': 101,
                 'No': 102,
                 'Lr': 103,
                 'Rf': 104,
                 'Db': 105,
                 'Sg': 106,
                 'Bh': 107,
                 'Hs': 108,
                 'Mt': 109,
                 'Ds': 110,
                 'Rg': 111,
                 'Cn': 112,
                 'Nh': 113,
                 'Fl': 114,
                 'Mc': 115,
                 'Lv': 116,
                 'Ts': 117,
                 'Og': 118}

# Define inverse mapping
_Z_to_element = {z: elt for elt, z in _element_to_Z.items()}


def element_to_Z(element: str) -> int:
    """
    Returns the atomic number Z associated with an elements 1-2 letter name.
    Returns the same integer if an integer is passed in.

    :param element:
    :return:
    """

    if element in _user_element_to_Z:
        return _user_element_to_Z[element]

    # If already integer, do nothing
    if isinstance(element, (int, np.integer)):
        return element
    if type(element).__module__ == 'numpy' and np.issubdtype(type(element),
                                                             np.integer):
        return element

    # If a string-casted integer, do nothing
    if isinstance(element, str) and element.isnumeric():
        return int(element)

    # Check that a valid element was passed in then return
    if _element_to_Z.get(element, None) is None:
        warn(f'Element as specified not found in list of element-Z mappings. '
             'If you would like to specify a custom element, use an integer '
             'of your choosing instead. Setting element {element} to integer '
             '0')
    return _element_to_Z.get(element, 0)


class NumpyEncoder(JSONEncoder):
    """
    Special json encoder for numpy types for serialization
    use as

    json.loads(... cls = NumpyEncoder)

    or:

    json.dumps(... cls = NumpyEncoder)

    Thanks to StackOverflow users karlB and fnunnari, who contributed this from:
    `https://stackoverflow.com/a/47626762`
    """

    def default(self, obj):
        """
        """
        if isinstance(obj, (np.int_, np.intc, np.intp, np.int8,
                            np.int16, np.int32, np.int64, np.uint8,
                            np.uint16, np.uint32, np.uint64)):
            return int(obj)
        elif isinstance(obj, (np.float_, np.float16, np.float32,
                              np.float64)):
            return float(obj)
        elif isinstance(obj, (np.ndarray,)):
            return obj.tolist()
        return JSONEncoder.default(self, obj)


def Z_to_element(Z: int) -> str:
    """
    Maps atomic numbers Z to element name, e.g. 1->"H".

    :param Z: Atomic number corresponding to element.
    :return: One or two-letter name of element.
    """

    if Z in _user_Z_to_element:
        return _user_Z_to_element[Z]

    # Check proper formatting
    if isinstance(Z, str):
        if Z.isnumeric():
            Z = int(Z)
        else:
            raise ValueError("Input Z is not a number. It should be an "
                             "integer")

    return _Z_to_element[Z]


_Z_to_mass = {1: 1.0079,
              2: 4.0026,
              3: 6.941,
              4: 9.0122,
              5: 10.811,
              6: 12.0107,
              7: 14.0067,
              8: 15.9994,
              9: 18.9984,
              10: 20.1797,
              11: 22.9897,
              12: 24.305,
              13: 26.9815,
              14: 28.0855,
              15: 30.9738,
              16: 32.065,
              17: 35.453,
              19: 39.0983,
              18: 39.948,
              20: 40.078,
              21: 44.9559,
              22: 47.867,
              23: 50.9415,
              24: 51.9961,
              25: 54.938,
              26: 55.845,
              28: 58.6934,
              27: 58.9332,
              29: 63.546,
              30: 65.39,
              31: 69.723,
              32: 72.64,
              33: 74.9216,
              34: 78.96,
              35: 79.904,
              36: 83.8,
              37: 85.4678,
              38: 87.62,
              39: 88.9059,
              40: 91.224,
              41: 92.9064,
              42: 95.94,
              43: 98,
              44: 101.07,
              45: 102.9055,
              46: 106.42,
              47: 107.8682,
              48: 112.411,
              49: 114.818,
              50: 118.71,
              51: 121.76,
              53: 126.9045,
              52: 127.6,
              54: 131.293,
              55: 132.9055,
              56: 137.327,
              57: 138.9055,
              58: 140.116,
              59: 140.9077,
              60: 144.24,
              61: 145,
              62: 150.36,
              63: 151.964,
              64: 157.25,
              65: 158.9253,
              66: 162.5,
              67: 164.9303,
              68: 167.259,
              69: 168.9342,
              70: 173.04,
              71: 174.967,
              72: 178.49,
              73: 180.9479,
              74: 183.84,
              75: 186.207,
              76: 190.23,
              77: 192.217,
              78: 195.078,
              79: 196.9665,
              80: 200.59,
              81: 204.3833,
              82: 207.2,
              83: 208.9804,
              84: 209,
              85: 210,
              86: 222,
              87: 223,
              88: 226,
              89: 227,
              91: 231.0359,
              90: 232.0381,
              93: 237,
              92: 238.0289,
              95: 243,
              94: 244,
              96: 247,
              97: 247,
              98: 251,
              99: 252,
              100: 257,
              101: 258,
              102: 259,
              104: 261,
              103: 262,
              105: 262,
              107: 264,
              106: 266,
              109: 268,
              111: 272,
              108: 277}



"""The :class:`AtomicEnvironment` object stores information about the local
environment of an atom. :class:`AtomicEnvironment` objects are inputs to the
2-, 3-, and 2+3-body kernels."""
import numpy as np
from copy import deepcopy
from math import ceil
from flare.struc import Structure
import flare.kernels.cutoffs as cf
from flare.utils.env_getarray import get_2_body_arrays, get_3_body_arrays, \
    get_m2_body_arrays



class AtomicEnvironment:
    """Contains information about the local environment of an atom,
    including arrays of pair and triplet distances and the chemical
    species of atoms in the environment.

    :param structure: Structure of atoms.
    :type structure: struc.Structure
    :param atom: Index of the atom in the structure.
    :type atom: int
    :param cutoffs: 2- and 3-body cutoff radii. 2-body if one cutoff is
    given, 2+3-body if two are passed.
    :type cutoffs: np.ndarray
    :param cutoffs_mask: a dictionary to store multiple cutoffs if neede
                         it should be exactly the same as the hyps mask
    :type cutoffs_mask: dict

    The cutoffs_mask allows the user to define multiple cutoffs for different
    bonds, triples, and many body interaction. This dictionary should be
    consistent with the hyps_mask used in the GuassianProcess object.

    * specie_mask: 118-long integer array descirbing which elements belong to
        like groups for determining which bond hyperparameters to use.
        For instance, [0,0,1,1,0 ...] assigns H to group 0, He and
        Li to group 1, and Be to group 0 (the 0th register is ignored).
    * nspecie: Integer, number of different species groups (equal to number of
        unique values in specie_mask).
    * ntwobody: Integer, number of different hyperparameter/cutoff sets to
        associate with different 2-body pairings of atoms in groups defined in
        specie_mask.
    * twobody_mask: Array of length nspecie^2, which describes the cutoff to
        associate with different pairings of species types. For example, if
        there are atoms of type 0 and 1, then twobody_mask defines which cutoff
        to use for parings [0-0, 0-1, 1-0, 1-1]: if we wanted cutoff0 for
        0-0 parings and set 1 for 0-1 and 1-1 pairings, then we would make
        twobody_mask [0, 1, 1, 1].
    * twobody_cutoff_list: Array of length ntwobody, which stores the cutoff
        used for different types of bonds defined in twobody_mask
    * ncut3b: Integer, number of different cutoffs sets to associate
        with different 3-body pariings of atoms in groups defined in
        specie_mask.
    * cut3b_mask: Array of length nspecie^2, which describes the cutoff to
        associate with different bond types in triplets. For example, in a
        triplet (C, O, H) , there are three cutoffs. Cutoffs for CH bond, CO
        bond and OH bond. If C and O are associate with atom group 1 in
        specie_mask and H are associate with group 0 in specie_mask, the
        cut3b_mask[1*nspecie+0] determines the C/O-H bond cutoff, and
        cut3b_mask[1*nspecie+1] determines the C-O bond cutoff. If we want the
        former one to use the 1st cutoff in threebody_cutoff_list and the later
        to use the 2nd cutoff in threebody_cutoff_list, the cut3b_mask should
        be [0, 0, 0, 1].
    * threebody_cutoff_list: Array of length ncut3b, which stores the cutoff
        used for different types of bonds in triplets.
    * nmanybody: Integer, number of different cutoffs set to associate with
        different coordination numbers.
    * manybody_mask: Similar to twobody_mask and cut3b_mask.
    * manybody_cutoff_list: Array of length nmanybody, stores the cutoff used
        for different many body terms

    Examples can be found at the end of in tests/test_env.py

    """

    all_kernel_types = ['twobody', 'threebody', 'manybody']
    ndim = {'twobody': 2, 'threebody': 3, 'manybody': 2, 'cut3b': 2}

    def __init__(self, structure: Structure, atom: int, cutoffs,
                 cutoffs_mask=None):

        self.structure = structure
        self.positions = structure.wrapped_positions
        self.cell = np.array(structure.cell)
        self.species = structure.coded_species

        # backward compatability
        if not isinstance(cutoffs, dict):
            newcutoffs = {'twobody': cutoffs[0]}
            if len(cutoffs) > 1:
                newcutoffs['threebody'] = cutoffs[1]
            if len(cutoffs) > 2:
                newcutoffs['manybody'] = cutoffs[2]
            cutoffs = newcutoffs

        if cutoffs_mask is None:
            cutoffs_mask = {'cutoffs': cutoffs}
        elif cutoffs is not None:
            cutoffs_mask['cutoffs'] = deepcopy(cutoffs)

        # Set the sweep array based on the max cutoff.
        sweep_val = ceil(np.max(list(cutoffs.values())) / structure.max_cutoff)
        self.sweep_val = sweep_val
        self.sweep_array = np.arange(-sweep_val, sweep_val + 1, 1)

        self.atom = atom
        self.ctype = structure.coded_species[atom]

        self.twobody_cutoff = 0
        self.threebody_cutoff = 0
        self.manybody_cutoff = 0

        self.ntwobody = 1
        self.ncut3b = 0
        self.nmanybody = 0

        self.nspecie = 1
        self.specie_mask = None
        self.twobody_mask = None
        self.threebody_mask = None
        self.manybody_mask = None
        self.twobody_cutoff_list = None
        self.threebody_cutoff_list = None
        self.manybody_cutoff_list = None

        self.setup_mask(cutoffs_mask)

        assert self.threebody_cutoff <= self.twobody_cutoff, \
            "2b cutoff has to be larger than 3b cutoff"
        # # TO DO, once the mb function is updated to use the bond_array_2
        # # this block should be activated.
        # assert self.manybody_cutoff <= self.twobody_cutoff, \
        #         "mb cutoff has to be larger than mb cutoff"

        self.bond_array_2 = None
        self.etypes = None
        self.bond_inds = None
        self.bond_array_3 = None
        self.cross_bond_inds = None
        self.cross_bond_dists = None
        self.triplet_counts = None
        self.q_array = None
        self.q_neigh_array = None
        self.q_grads = None
        self.q_neigh_grads = None
        self.unique_species = None
        self.etypes_mb = None

        self.compute_env()

    def setup_mask(self, cutoffs_mask):

        self.cutoffs_mask = deepcopy(cutoffs_mask)
        self.cutoffs = cutoffs_mask['cutoffs']

        for kernel in AtomicEnvironment.all_kernel_types:
            ndim = AtomicEnvironment.ndim[kernel]
            if kernel in self.cutoffs:
                setattr(self, kernel + '_cutoff', self.cutoffs[kernel])

        if (self.twobody_cutoff == 0):
            self.twobody_cutoff = \
                np.max([self.threebody_cutoff, self.manybody_cutoff])

            self.cutoffs['twobody'] = self.twobody_cutoff

        self.nspecie = cutoffs_mask.get('nspecie', 1)
        if 'specie_mask' in cutoffs_mask:
            self.specie_mask = np.array(cutoffs_mask['specie_mask'],
                                        dtype=np.int)

        for kernel in AtomicEnvironment.all_kernel_types:
            ndim = AtomicEnvironment.ndim[kernel]
            if kernel in self.cutoffs:
                setattr(self, kernel + '_cutoff', self.cutoffs[kernel])
                setattr(self, 'n' + kernel, 1)
                if kernel != 'threebody':
                    name_list = [kernel + '_cutoff_list',
                                 'n' + kernel, kernel + '_mask']
                    for name in name_list:
                        if name in cutoffs_mask:
                            setattr(self, name, cutoffs_mask[name])
                else:
                    self.ncut3b = cutoffs_mask.get('ncut3b', 1)
                    self.cut3b_mask = cutoffs_mask.get('cut3b_mask', None)
                    if 'threebody_cutoff_list' in cutoffs_mask:
                        self.threebody_cutoff_list = \
                            np.array(cutoffs_mask['threebody_cutoff_list'],
                                     dtype=np.float)


    def compute_env(self):

        # get 2-body arrays
        if self.ntwobody >= 1:
            bond_array_2, bond_positions_2, etypes, bond_inds = \
                get_2_body_arrays(
                    self.positions, self.atom, self.cell, self.twobody_cutoff,
                    self.twobody_cutoff_list, self.species, self.sweep_array,
                    self.nspecie, self.specie_mask, self.twobody_mask)

            self.bond_array_2 = bond_array_2
            self.etypes = etypes
            self.bond_inds = bond_inds

        # if 2 cutoffs are given, create 3-body arrays
        if self.ncut3b > 0:
            bond_array_3, cross_bond_inds, cross_bond_dists, triplet_counts = \
                get_3_body_arrays(
                    bond_array_2, bond_positions_2, self.species[self.atom],
                    etypes, self.threebody_cutoff, self.threebody_cutoff_list,
                    self.nspecie, self.specie_mask, self.cut3b_mask)

            self.bond_array_3 = bond_array_3
            self.cross_bond_inds = cross_bond_inds
            self.cross_bond_dists = cross_bond_dists
            self.triplet_counts = triplet_counts

        # if 3 cutoffs are given, create many-body arrays
        if self.nmanybody > 0:
            self.q_array, self.q_neigh_array, self.q_grads, \
                self.q_neigh_grads, self.unique_species, self.etypes_mb = \
                get_m2_body_arrays(
                    self.positions, self.atom, self.cell, self.manybody_cutoff,
                    self.manybody_cutoff_list, self.species, self.sweep_array,
                    self.nspecie, self.specie_mask, self.manybody_mask,
                    cf.quadratic_cutoff)

    def as_dict(self, include_structure: bool = False):
        """
        Returns Atomic Environment object as a dictionary for serialization
        purposes. Optional to not include the structure to avoid redundant
        information.
        :return:
        """

        dictionary = dict(vars(self))
        dictionary['object'] = 'AtomicEnvironment'
        dictionary['forces'] = self.structure.forces

        # Backward compatibility for older models: Cutoffs mask.
        # Can be deleted one day if support is dropped for older (Pre June
        # 2020) pickled environment objects.
        cutoffs_mask = getattr(self, 'cutoffs_mask', {'cutoffs': self.cutoffs})
        if not hasattr(self, 'cutoffs_mask'):
            self.cutoffs_mask = cutoffs_mask
        dictionary['cutoffs_mask'] = cutoffs_mask

        if not include_structure:
            del dictionary['structure']
        else:
            dictionary['structure'] = dictionary['structure'].as_dict()

        return dictionary

    @staticmethod
    def from_dict(dictionary):
        """
        Loads in atomic environment object from a dictionary which was
        serialized by the to_dict method.

        :param dictionary: Dictionary describing atomic environment.
        """
        # TODO Instead of re-computing 2 and 3 body environment,
        # directly load in, this would be much more efficient

        struc = Structure(cell=np.array(dictionary['cell']),
                          positions=dictionary['positions'],
                          species=dictionary['species'])
        index = dictionary['atom']

        if dictionary.get('cutoffs') is not None:
            cutoffs = dictionary['cutoffs']
        else:
            cutoffs = {}

        cutoffs_mask = dictionary.get('cutoffs_mask', None)

        return AtomicEnvironment(
            struc, index, cutoffs, cutoffs_mask=cutoffs_mask)

    def __str__(self):
        atom_type = self.ctype
        neighbor_types = self.etypes
        n_neighbors = len(self.bond_array_2)
        string = 'Atomic Env. of Type {} surrounded by {} atoms ' \
                 'of Types {}'.format(atom_type, n_neighbors,
                                      sorted(list(set(neighbor_types))))

        return string



import sys, subprocess
import numpy as np

from copy import deepcopy
from typing import List, Tuple
from flare import gp, env, struc, otf
from flare.gp import GaussianProcess


class OtfAnalysis:
    def __init__(self, filename, calculate_energy=False):
        self.filename = filename

        self.calculate_energy = calculate_energy

        blocks = split_blocks(filename)

        self.header = parse_header_information(blocks[0])
        self.noa = self.header["atoms"]
        self.noh = self.header["n_hyps"]

        self.position_list = []
        self.cell_list = []
        self.force_list = []
        self.stress_list = []
        self.uncertainty_list = []
        self.velocity_list = []
        self.temperatures = []
        self.dft_frames = []
        self.dft_times = []
        self.times = []
        self.msds = []
        self.energies = []
        self.thermostat = {}

        self.gp_position_list = []
        self.gp_cell_list = []
        self.gp_force_list = []
        self.gp_stress_list = []
        self.gp_uncertainty_list = []
        self.gp_velocity_list = []
        self.gp_atom_list = []
        self.gp_species_list = []
        self.gp_atom_count = []
        self.gp_thermostat = {}

        self.gp_hyp_list = [self.header["hyps"]]

        self.mae_list = []
        self.maf_list = []

        self.parse_pos_otf(blocks[1:])

        if self.calculate_energy:
            self.energies = energies

    def make_gp(
        self, cell=None, call_no=None, hyps=None, init_gp=None, hyp_no=None, **kwargs,
    ):

        if "restart" in self.header and self.header["restart"] > 0:
            assert init_gp is not None, "Please input the init_gp as the gp model dumpped"\
                    "before restarting otf."

        if call_no is None:
            call_no = len(self.gp_position_list)
        if hyp_no is None:
            hyp_no = len(self.gp_hyp_list)  # use the last hyps by default
        if hyps is None:
            # check out the last non-empty element from the list
            hyps = self.gp_hyp_list[hyp_no - 1]
        if cell is None:
            cell = self.header["cell"]

        if init_gp is None:
            # Use run's values as extracted from header
            # TODO Allow for kernel gradient in header

            dictionary = deepcopy(self.header)
            dictionary["hyps"] = hyps
            for k in kwargs:
                if kwargs[k] is not None:
                    dictionary[k] = kwargs[k]

            gp_model = GaussianProcess.from_dict(dictionary)
        else:
            gp_model = init_gp
            gp_model.hyps = hyps

        for (positions, forces, atoms, species) in zip(
            self.gp_position_list[:call_no],
            self.gp_force_list[:call_no],
            self.gp_atom_list[:call_no],
            self.gp_species_list[:call_no],
        ):

            struc_curr = struc.Structure(cell, species, positions)

            gp_model.update_db(struc_curr, forces, custom_range=atoms)

        gp_model.set_L_alpha()

        return gp_model

    @staticmethod
    def get_gp_activation(gp_model):
        pass

    def parse_pos_otf(self, blocks):
        """
        Exclusively parses MD run information
        :param filename:
        :return:
        """
        n_steps = len(blocks) - 1

        for block in blocks:
            for index, line in enumerate(block):
                # DFT frame
                if line.startswith("*-Frame"):
                    dft_frame_line = line.split()
                    self.dft_frames.append(int(dft_frame_line[1]))
                    dft_time_line = block[index + 1].split()
                    self.dft_times.append(float(dft_time_line[-2]))

                    # TODO: generalize this to account for arbitrary starting list
                    append_atom_lists(
                        self.gp_species_list,
                        self.gp_position_list,
                        self.gp_force_list,
                        self.gp_uncertainty_list,
                        self.gp_velocity_list,
                        block,
                        index,
                        self.noa,
                        True,
                        self.noh,
                    )

                    post_frame = block[index + 3 + self.noa :]
                    extract_global_info(
                        self.gp_cell_list,
                        self.gp_stress_list,
                        self.gp_thermostat,
                        post_frame,
                    )

                    extract_gp_info(
                        post_frame,
                        self.mae_list,
                        self.maf_list,
                        self.gp_atom_list,
                        self.gp_hyp_list,
                        self.noh,
                    )

                # MD frame
                if line.startswith("-Frame"):
                    n_steps += 1
                    time_line = block[index + 1].split()
                    sim_time = float(time_line[2])
                    self.times.append(sim_time)

                    # TODO: generalize this to account for arbitrary starting list
                    append_atom_lists(
                        [],
                        self.position_list,
                        self.force_list,
                        self.uncertainty_list,
                        self.velocity_list,
                        block,
                        index,
                        self.noa,
                        False,
                        self.noh,
                    )

                    post_frame = block[index + 3 + self.noa :]
                    extract_global_info(
                        self.cell_list, self.stress_list, self.thermostat, post_frame,
                    )

                    self.msds.append(
                        np.mean((self.position_list[-1] - self.position_list[0]) ** 2)
                    )

    def output_md_structures(self):
        """
        Returns structure objects corresponding to the MD frames of an OTF run.
        :return:
        """

        positions = self.position_list
        structures = []
        cell = self.header["cell"]
        species = self.header["species"]
        forces = self.force_list
        stds = self.uncertainty_list
        for i in range(len(positions)):
            cur_struc = struc.Structure(
                cell=cell, species=species, positions=positions[i]
            )
            cur_struc.forces = forces[i]
            cur_struc.stds = stds[i]
            structures.append(cur_struc)
        return structures


def split_blocks(filename):
    with open(filename, "r") as f:
        lines = f.readlines()
        head = 0
        blocks = []
        for index, line in enumerate(lines):
            if line.startswith("---"):
                blocks.append(lines[head:index])
                head = index + 1
    return blocks


def parse_header_information(lines) -> dict:
    """
    Get information about the run from the header of the file
    :param outfile:
    :return:
    """
    header_info = {}

    cutoffs_dict = {}
    for i, line in enumerate(lines):
        line_lower = line.lower()

        # gp related
        if "cutoffs" in line_lower:
            if "{" in line_lower:
                line = line[len("Cutoffs: ") :]
                cutoffs = eval(line)
            else:
                line = line.split(":")[1].strip()
                line = line.strip("[").strip("]")
                line = line.split()
                cutoffs = []
                for val in line:
                    try:
                        cutoffs.append(float(val))
                    except:
                        cutoffs.append(float(val[:-1]))
            header_info["cutoffs"] = cutoffs

        if "number of hyperparameters" in line_lower:
            n_hyps = int(line.split(":")[1].strip())
            header_info["n_hyps"] = n_hyps 
            
            # parse hyps
            new_line = lines[i+1].replace("[", "")
            new_line = new_line.replace("]", "")
            assert "hyperparameter" in new_line.lower()

            hyps_array = new_line[new_line.find(":")+1:].split()
            if len(hyps_array) < n_hyps:
                next_new_line = lines[i+2].replace("[", "")
                next_new_line = next_new_line.replace("]", "")
                hyps_array += next_new_line.split()
            assert len(hyps_array) == n_hyps

            hyps_array = [float(h.strip()) for h in hyps_array]
            header_info["hyps"] = np.array(hyps_array)

        if "kernel_name" in line_lower:
            header_info["kernel_name"] = line.split(":")[1].strip()
        elif "kernels" in line_lower:
            line = line.split(":")[1].strip()
            line = line.strip("[").strip("]")
            line = line.split()
            header_info["kernels"] = line
        elif "kernel" in line_lower:
            header_info["kernel_name"] = line.split(":")[1].strip()

        for kw in header_dict:
            get_header_item(line, header_info, kw)

        if "optimization algorithm" in line:
            header_info["algo"] = str(line.split(":")[1].strip()).upper()

        if "system species" in line_lower:
            line = line.split(":")[1]
            line = line.split("'")
            species = [item for item in line if item.isalpha()]
            header_info["species_set"] = set(species)
        if "periodic cell" in line_lower:
            vectors = []
            for cell_line in lines[i + 1 : i + 4]:
                cell_line = cell_line.strip().replace("[", "").replace("]", "")
                vec = cell_line.split()
                vector = [float(vec[0]), float(vec[1]), float(vec[2])]
                vectors.append(vector)
            header_info["cell"] = np.array(vectors)
        if "previous positions" in line_lower:
            struc_spec = []
            prev_positions = []
            for pos_line in lines[i + 1 : i + 1 + header_info.get("atoms", 0)]:
                pos = pos_line.split()
                struc_spec.append(pos[0])
                prev_positions.append((float(pos[1]), float(pos[2]), float(pos[3])))
            header_info["species"] = struc_spec
            header_info["prev_positions"] = np.array(prev_positions)

    return header_info


def get_header_item(line, header_info, kw):
    if not isinstance(line, str):
        return

    pattern = header_dict[kw][0]
    value_type = header_dict[kw][1]

    if header_dict[kw][2]:
        pattern = pattern.lower()
        line = line.lower()

    if pattern in line:
        header_info[kw] = value_type(line.split(":")[1].strip())


header_dict = {
    "restart": ["Restart", int, False],
    "frames": ["Frames", int, True],
    "atoms": ["Number of atoms", int, True],
    "dt": ["Timestep", float, True],
}


def append_atom_lists(
    species_list: List[str],
    position_list: List[np.ndarray],
    force_list: List[np.ndarray],
    uncertainty_list: List[np.ndarray],
    velocity_list: List[np.ndarray],
    lines: List[str],
    index: int,
    noa: int,
    dft_call: bool,
    noh: int,
) -> None:

    """Update lists containing atom information at each snapshot."""

    species, positions, forces, uncertainties, velocities = parse_snapshot(
        lines, index, noa, dft_call, noh
    )

    species_list.append(species)
    position_list.append(positions)
    force_list.append(forces)
    uncertainty_list.append(uncertainties)
    velocity_list.append(velocities)


def parse_snapshot(lines, index, noa, dft_call, noh):
    """Parses snapshot of otf output file."""

    # initialize values
    species = []
    positions = np.zeros((noa, 3))
    forces = np.zeros((noa, 3))
    uncertainties = np.zeros((noa, 3))
    velocities = np.zeros((noa, 3))

    # Current setting for # of lines to skip after Frame marker
    skip = 3

    for count, frame_line in enumerate(lines[(index + skip) : (index + skip + noa)]):
        # parse frame line
        spec, position, force, uncertainty, velocity = parse_frame_line(frame_line)

        # update values
        species.append(spec)
        positions[count] = position
        forces[count] = force
        uncertainties[count] = uncertainty
        velocities[count] = velocity

    return species, positions, forces, uncertainties, velocities


def strip_and_split(line):
    """
    Helper function which saves a few lines of code elsewhere
    :param line:
    :return:
    """

    line = line.strip().split()
    stripped_line = [subline.strip() for subline in line]

    return stripped_line


def parse_frame_line(frame_line):
    """parse a line in otf output.
    :param frame_line: frame line to be parsed
    :type frame_line: string
    :return: species, position, force, uncertainty, and velocity of atom
    :rtype: list, np.arrays
    """

    frame_line = frame_line.split()

    spec = str(frame_line[0])
    position = np.array([float(n) for n in frame_line[1:4]])
    force = np.array([float(n) for n in frame_line[4:7]])
    uncertainty = np.array([float(n) for n in frame_line[7:10]])
    velocity = np.array([float(n) for n in frame_line[10:13]])

    return spec, position, force, uncertainty, velocity


def extract_global_info(
    cell_list, stress_list, thermostat, block,
):

    for ind, line in enumerate(block):
        if "cell" in line:
            vectors = []
            for cell_line in block[ind + 1 : ind + 4]:
                cell_line = cell_line.strip().replace("[", "").replace("]", "")
                vec = cell_line.split()
                vector = [float(vec[0]), float(vec[1]), float(vec[2])]
                vectors.append(vector)
            cell_list.append(vectors)
        if "Stress" in line:
            vectors = []
            stress_line = block[ind + 2].split()
            vectors = [float(s) for s in stress_line]
            stress_list.append(vectors)

        for t in [
            "Pressure",
            "Temperature",
            "Kinetic energy",
            "Potential energy",
            "Total energy",
        ]:
            get_thermostat(thermostat, t, line)


def get_thermostat(thermostat, kw, line):
    if kw in line:
        value = float(line.split()[-1])
        kw = kw.lower()
        if kw in thermostat:
            thermostat[kw].append(value)
        else:
            thermostat[kw] = [value]


def extract_gp_info(block, mae_list, maf_list, atoms_list, hyps_list, noh):
    """
    Exclusively parses DFT run information
    :param filename:
    :return:
    """
    for ind, line in enumerate(block):
        if "mean absolute error" in line:
            value = float(line.split()[3])
            mae_list.append(value)
        if "mean absolute dft component" in line:
            value = float(line.split()[4])
            maf_list.append(value)

        # keep track of atom number
        if line.startswith("Adding atom"):
            atoms_added = []
            line_split = line.split()
            atom_strings = line_split[2:-4]
            for n, atom_string in enumerate(atom_strings):
                if n == 0:
                    atoms_added.append(int(atom_string[1:-1]))
                else:
                    atoms_added.append(int(atom_string[0:-1]))
            atoms_list.append(atoms_added)

        # keep track of hyperparameters
        if line.startswith("GP hyperparameters:"):
            hyps = []
            for hyp_line in block[(ind + 1) : (ind + 1 + noh)]:
                hyp_line = hyp_line.split()
                hyps.append(float(hyp_line[-1]))
            hyps = np.array(hyps)
            hyps_list.append(hyps)



import numpy as np
from math import exp
from numba import njit
import flare.kernels.cutoffs as cf

# -----------------------------------------------------------------------------
#                            general helper functions
# -----------------------------------------------------------------------------


@njit
def grad_constants(sig, ls):
    sig2 = sig * sig
    sig3 = 2 * sig

    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2
    ls4 = 1 / (ls * ls * ls)
    ls5 = ls * ls
    ls6 = ls2 * ls4

    return sig2, sig3, ls1, ls2, ls3, ls4, ls5, ls6


@njit
def force_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, sig2):
    """Helper function for computing the force/force kernel between two
    pairs or triplets of atoms of the same type.

    See Table IV of the SI of the FLARE paper for definitions of intermediate
    quantities.

    Returns:
        float: Force/force kernel between two pairs or triplets of atoms of
            the same type.
    """
    E = exp(-D * ls1)
    I = fdi * fdj
    J = B * ls2 * fi * fdj
    K = - C * ls2 * fdi * fj
    L = (A * ls2 - B * C * ls3) * fi * fj
    M = sig2 * (I + J + K + L) * E
    return M


@njit
def grad_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, ls4, ls5, ls6,
                sig2, sig3):
    E = exp(-D * ls1)
    F = E * B * ls2
    G = -E * C * ls2
    H = A * E * ls2 - B * C * E * ls3
    I = E * fdi * fdj
    J = F * fi * fdj
    K = G * fdi * fj
    L = H * fi * fj
    M = I + J + K + L
    N = sig2 * M
    O = sig3 * M
    P = D * ls4
    Q = B * (ls2 * P - 2 * ls4)
    R = -C * (ls2 * P - 2 * ls4)
    S = (A * ls5 - B * C) * (P * ls3 - 4 * ls6) + 2 * A * ls4
    T = P * fdi * fdj
    U = Q * fi * fdj
    V = R * fdi * fj
    W = S * fi * fj
    X = sig2 * (T + U + V + W) * E

    return N, O, X


@njit
def force_energy_helper(B, D, fi, fj, fdi, ls1, ls2, sig2):
    E = exp(-D * ls1)
    F = B * ls2
    G = -F * fi * fj
    H = -fdi * fj
    I = sig2 * (G + H) * E

    return I


@njit
def three_body_helper_1(ci1, ci2, cj1, cj2, r11, r22, r33, fi, fj, fdi, fdj,
                        ls1, ls2, ls3, sig2):
    A = ci1 * cj1 + ci2 * cj2
    B = r11 * ci1 + r22 * ci2
    C = r11 * cj1 + r22 * cj2
    D = r11 * r11 + r22 * r22 + r33 * r33

    M = force_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, sig2)

    return M


@njit
def three_body_helper_2(ci1, ci2, cj1, cj2, r12, r23, r31, fi, fj, fdi, fdj,
                        ls1, ls2, ls3, sig2):
    A = ci1 * cj2
    B = r12 * ci1 + r23 * ci2
    C = r12 * cj2 + r31 * cj1
    D = r12 * r12 + r23 * r23 + r31 * r31

    M = force_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, sig2)

    return M


@njit
def three_body_sf_1(ci1, ci2, cj1, cj2, r11, r22, r33, fi, fj, fdi, fdj,
                    ls1, ls2, ls3, sig2, coord1, coord2, fdi1, fdi2):
    A = ci1 * cj1 * coord1 + ci2 * cj2 * coord2
    B = r11 * ci1 * coord1 + r22 * ci2 * coord2
    C = r11 * cj1 + r22 * cj2
    D = r11 * r11 + r22 * r22 + r33 * r33
    E = exp(-D * ls1)
    F = B * ls2
    G = -C * ls2
    H = A * ls2 - B * C * ls3
    I = (fdi1 * coord1 + fdi2 * coord2) * fdj
    J = F * fi * fdj
    K = G * (fdi1 * coord1 + fdi2 * coord2) * fj
    L = H * fi * fj
    M = -sig2 * (I + J + K + L) * E

    return M


@njit
def three_body_sf_2(ci1, ci2, cj1, cj2, r12, r23, r31, fi, fj, fdi, fdj,
                    ls1, ls2, ls3, sig2, coord1, coord2, fdi1, fdi2):
    A = ci1 * cj2 * coord1
    B = r12 * ci1 * coord1 + r23 * ci2 * coord2
    C = r12 * cj2 + r31 * cj1
    D = r12 * r12 + r23 * r23 + r31 * r31
    E = exp(-D * ls1)
    F = B * ls2
    G = -C * ls2
    H = A * ls2 - B * C * ls3
    I = (fdi1 * coord1 + fdi2 * coord2) * fdj
    J = F * fi * fdj
    K = G * (fdi1 * coord1 + fdi2 * coord2) * fj
    L = H * fi * fj
    M = -sig2 * (I + J + K + L) * E

    return M


@njit
def three_body_ss_1(ci1, ci2, cj1, cj2, r11, r22, r33, fi, fj, fdi, fdj,
                    ls1, ls2, ls3, sig2, coord1, coord2, coord3, coord4,
                    fdi_p1, fdi_p2, fdj_p1, fdj_p2):
    A = ci1 * cj1 * coord1 * coord3 + ci2 * cj2 * coord2 * coord4
    B = r11 * ci1 * coord1 + r22 * ci2 * coord2
    C = r11 * cj1 * coord3 + r22 * cj2 * coord4
    D = r11 * r11 + r22 * r22 + r33 * r33
    E = exp(-D * ls1)
    F = B * ls2
    G = -C * ls2
    H = A * ls2 - B * C * ls3
    I = (fdi_p1 * coord1 + fdi_p2 * coord2) * \
        (fdj_p1 * coord3 + fdj_p2 * coord4)
    J = F * fi * (fdj_p1 * coord3 + fdj_p2 * coord4)
    K = G * (fdi_p1 * coord1 + fdi_p2 * coord2) * fj
    L = H * fi * fj
    M = sig2 * (I + J + K + L) * E

    return M


@njit
def three_body_ss_2(ci1, ci2, cj1, cj2, r12, r23, r31, fi, fj, fdi, fdj,
                    ls1, ls2, ls3, sig2, coord1, coord2, coord3, coord4,
                    fdi_p1, fdi_p2, fdj_p1, fdj_p2):
    A = ci1 * cj2 * coord1 * coord4
    B = r12 * ci1 * coord1 + r23 * ci2 * coord2
    C = r12 * cj2 * coord4 + r31 * cj1 * coord3
    D = r12 * r12 + r23 * r23 + r31 * r31
    E = exp(-D * ls1)
    F = B * ls2
    G = -C * ls2
    H = A * ls2 - B * C * ls3
    I = (fdi_p1 * coord1 + fdi_p2 * coord2) * \
        (fdj_p1 * coord3 + fdj_p2 * coord4)
    J = F * fi * (fdj_p1 * coord3 + fdj_p2 * coord4)
    K = G * (fdi_p1 * coord1 + fdi_p2 * coord2) * fj
    L = H * fi * fj
    M = sig2 * (I + J + K + L) * E

    return M


@njit
def three_body_grad_helper_1(ci1, ci2, cj1, cj2, r11, r22, r33, fi, fj, fdi,
                             fdj, ls1, ls2, ls3, ls4, ls5, ls6, sig2, sig3):
    A = ci1 * cj1 + ci2 * cj2
    B = r11 * ci1 + r22 * ci2
    C = r11 * cj1 + r22 * cj2
    D = r11 * r11 + r22 * r22 + r33 * r33

    N, O, X = grad_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, ls4,
                          ls5, ls6, sig2, sig3)

    return N, O, X


@njit
def three_body_grad_helper_2(ci1, ci2, cj1, cj2, r12, r23, r31, fi, fj, fdi,
                             fdj, ls1, ls2, ls3, ls4, ls5, ls6, sig2, sig3):
    A = ci1 * cj2
    B = r12 * ci1 + r23 * ci2
    C = r12 * cj2 + r31 * cj1
    D = r12 * r12 + r23 * r23 + r31 * r31

    N, O, X = grad_helper(A, B, C, D, fi, fj, fdi, fdj, ls1, ls2, ls3, ls4,
                          ls5, ls6, sig2, sig3)

    return N, O, X


@njit
def three_body_en_helper(ci1, ci2, r11, r22, r33, fi, fj, fdi, ls1, ls2, sig2):
    B = r11 * ci1 + r22 * ci2
    D = r11 * r11 + r22 * r22 + r33 * r33

    return force_energy_helper(B, D, fi, fj, fdi, ls1, ls2, sig2)


@njit
def three_body_ee_perm(r11, r12, r13, r21, r22, r23, r31, r32, r33, c1, c2,
                       ei1, ei2, ej1, ej2, fi, fj, ls2, sig2):
    kern = 0

    if (c1 == c2):
        if (ei1 == ej1) and (ei2 == ej2):
            C1 = r11 * r11 + r22 * r22 + r33 * r33
            kern += exp(-C1 * ls2)
        if (ei1 == ej2) and (ei2 == ej1):
            C3 = r12 * r12 + r21 * r21 + r33 * r33
            kern += exp(-C3 * ls2)
    if (c1 == ej1):
        if (ei1 == ej2) and (ei2 == c2):
            C5 = r13 * r13 + r21 * r21 + r32 * r32
            kern += exp(-C5 * ls2)
        if (ei1 == c2) and (ei2 == ej2):
            C2 = r11 * r11 + r23 * r23 + r32 * r32
            kern += exp(-C2 * ls2)
    if (c1 == ej2):
        if (ei1 == ej1) and (ei2 == c2):
            C6 = r13 * r13 + r22 * r22 + r31 * r31
            kern += exp(-C6 * ls2)
        if (ei1 == c2) and (ei2 == ej1):
            C4 = r12 * r12 + r23 * r23 + r31 * r31
            kern += exp(-C4 * ls2)

    return kern * sig2 * fi * fj


@njit
def three_body_fe_perm(r11, r12, r13, r21, r22, r23, r31, r32, r33, c1, c2,
                       ci1, ci2, ei1, ei2, ej1, ej2, fi, fj, fdi, ls1, ls2,
                       sig2):
    kern = 0

    if (c1 == c2):
        if (ei1 == ej1) and (ei2 == ej2):
            kern += three_body_en_helper(
                ci1, ci2, r11, r22, r33, fi, fj, fdi, ls1, ls2, sig2)
        if (ei1 == ej2) and (ei2 == ej1):
            kern += three_body_en_helper(
                ci1, ci2, r12, r21, r33, fi, fj, fdi, ls1, ls2, sig2)
    if (c1 == ej1):
        if (ei1 == ej2) and (ei2 == c2):
            kern += three_body_en_helper(
                ci1, ci2, r13, r21, r32, fi, fj, fdi, ls1, ls2, sig2)
        if (ei1 == c2) and (ei2 == ej2):
            kern += three_body_en_helper(
                ci1, ci2, r11, r23, r32, fi, fj, fdi, ls1, ls2, sig2)
    if (c1 == ej2):
        if (ei1 == ej1) and (ei2 == c2):
            kern += three_body_en_helper(
                ci1, ci2, r13, r22, r31, fi, fj, fdi, ls1, ls2, sig2)
        if (ei1 == c2) and (ei2 == ej1):
            kern += three_body_en_helper(
                ci1, ci2, r12, r23, r31, fi, fj, fdi, ls1, ls2, sig2)

    return kern


@njit
def three_body_ff_perm(r11, r12, r13, r21, r22, r23, r31, r32, r33, c1, c2,
                       ci1, ci2, cj1, cj2, ei1, ei2, ej1, ej2, fi, fj, fdi,
                       fdj, ls1, ls2, ls3, sig2):

    kern = 0

    if (c1 == c2):
        if (ei1 == ej1) and (ei2 == ej2):
            kern += \
                three_body_helper_1(ci1, ci2, cj1, cj2, r11, r22, r33, fi,
                                    fj, fdi, fdj, ls1, ls2, ls3, sig2)
        if (ei1 == ej2) and (ei2 == ej1):
            kern += \
                three_body_helper_1(ci1, ci2, cj2, cj1, r12, r21, r33, fi,
                                    fj, fdi, fdj, ls1, ls2, ls3, sig2)
    if (c1 == ej1):
        if (ei1 == ej2) and (ei2 == c2):
            kern += \
                three_body_helper_2(ci2, ci1, cj2, cj1, r21, r13, r32, fi,
                                    fj, fdi, fdj, ls1, ls2, ls3, sig2)
        if (ei1 == c2) and (ei2 == ej2):
            kern += \
                three_body_helper_2(ci1, ci2, cj2, cj1, r11, r23, r32, fi,
                                    fj, fdi, fdj, ls1, ls2, ls3, sig2)
    if (c1 == ej2):
        if (ei1 == ej1) and (ei2 == c2):
            kern += \
                three_body_helper_2(ci2, ci1, cj1, cj2, r22, r13, r31, fi,
                                    fj, fdi, fdj, ls1, ls2, ls3, sig2)
        if (ei1 == c2) and (ei2 == ej1):
            kern += \
                three_body_helper_2(ci1, ci2, cj1, cj2, r12, r23, r31, fi,
                                    fj, fdi, fdj, ls1, ls2, ls3, sig2)

    return kern


@njit
def three_body_sf_perm(r11, r12, r13, r21, r22, r23, r31, r32, r33, c1, c2,
                       ci1, ci2, cj1, cj2, ei1, ei2, ej1, ej2, fi, fj, fdi,
                       fdj, ls1, ls2, ls3, sig2, coord1, coord2, fdi1, fdi2):

    kern = 0

    if (c1 == c2):
        if (ei1 == ej1) and (ei2 == ej2):
            kern += \
                three_body_sf_1(ci1, ci2, cj1, cj2, r11, r22, r33, fi,
                                fj, fdi, fdj, ls1, ls2, ls3, sig2,
                                coord1, coord2, fdi1, fdi2)
        if (ei1 == ej2) and (ei2 == ej1):
            kern += \
                three_body_sf_1(ci1, ci2, cj2, cj1, r12, r21, r33, fi,
                                fj, fdi, fdj, ls1, ls2, ls3, sig2,
                                coord1, coord2, fdi1, fdi2)
    if (c1 == ej1):
        if (ei1 == ej2) and (ei2 == c2):
            kern += \
                three_body_sf_2(ci2, ci1, cj2, cj1, r21, r13, r32, fi,
                                fj, fdi, fdj, ls1, ls2, ls3, sig2,
                                coord2, coord1, fdi2, fdi1)
        if (ei1 == c2) and (ei2 == ej2):
            kern += \
                three_body_sf_2(ci1, ci2, cj2, cj1, r11, r23, r32, fi,
                                fj, fdi, fdj, ls1, ls2, ls3, sig2,
                                coord1, coord2, fdi1, fdi2)
    if (c1 == ej2):
        if (ei1 == ej1) and (ei2 == c2):
            kern += \
                three_body_sf_2(ci2, ci1, cj1, cj2, r22, r13, r31, fi,
                                fj, fdi, fdj, ls1, ls2, ls3, sig2,
                                coord2, coord1, fdi2, fdi1)
        if (ei1 == c2) and (ei2 == ej1):
            kern += \
                three_body_sf_2(ci1, ci2, cj1, cj2, r12, r23, r31, fi,
                                fj, fdi, fdj, ls1, ls2, ls3, sig2,
                                coord1, coord2, fdi1, fdi2)

    return kern


@njit
def three_body_ss_perm(r11, r12, r13, r21, r22, r23, r31, r32, r33, c1, c2,
                       ci1, ci2, cj1, cj2, ei1, ei2, ej1, ej2, fi, fj, fdi,
                       fdj, ls1, ls2, ls3, sig2, coord1, coord2, coord3,
                       coord4, fdi_p1, fdi_p2, fdj_p1, fdj_p2):

    kern = 0

    if (c1 == c2):
        if (ei1 == ej1) and (ei2 == ej2):
            kern += \
                three_body_ss_1(ci1, ci2, cj1, cj2, r11, r22, r33, fi,
                                fj, fdi, fdj, ls1, ls2, ls3, sig2,
                                coord1, coord2, coord3, coord4, fdi_p1,
                                fdi_p2, fdj_p1, fdj_p2)
        if (ei1 == ej2) and (ei2 == ej1):
            kern += \
                three_body_ss_1(ci1, ci2, cj2, cj1, r12, r21, r33, fi,
                                fj, fdi, fdj, ls1, ls2, ls3, sig2,
                                coord1, coord2, coord4, coord3, fdi_p1,
                                fdi_p2, fdj_p2, fdj_p1)
    if (c1 == ej1):
        if (ei1 == ej2) and (ei2 == c2):
            kern += \
                three_body_ss_2(ci2, ci1, cj2, cj1, r21, r13, r32, fi,
                                fj, fdi, fdj, ls1, ls2, ls3, sig2,
                                coord2, coord1, coord4, coord3, fdi_p2,
                                fdi_p1, fdj_p2, fdj_p1)
        if (ei1 == c2) and (ei2 == ej2):
            kern += \
                three_body_ss_2(ci1, ci2, cj2, cj1, r11, r23, r32, fi,
                                fj, fdi, fdj, ls1, ls2, ls3, sig2,
                                coord1, coord2, coord4, coord3, fdi_p1,
                                fdi_p2, fdj_p2, fdj_p1)
    if (c1 == ej2):
        if (ei1 == ej1) and (ei2 == c2):
            kern += \
                three_body_ss_2(ci2, ci1, cj1, cj2, r22, r13, r31, fi,
                                fj, fdi, fdj, ls1, ls2, ls3, sig2,
                                coord2, coord1, coord3, coord4, fdi_p2,
                                fdi_p1, fdj_p1, fdj_p2)
        if (ei1 == c2) and (ei2 == ej1):
            kern += \
                three_body_ss_2(ci1, ci2, cj1, cj2, r12, r23, r31, fi,
                                fj, fdi, fdj, ls1, ls2, ls3, sig2,
                                coord1, coord2, coord3, coord4, fdi_p1,
                                fdi_p2, fdj_p1, fdj_p2)

    return kern


@njit
def three_body_se_helper(ci1, ci2, r11, r22, r33, fi, fj, fdi, ls1, ls2, sig2,
                         coord1, coord2, fdi1, fdi2):
    p1 = r11 * r11 + r22 * r22 + r33 * r33
    p2 = exp(-p1 * ls1)
    p3 = p2 * ls2 * fi * fj
    p4 = p2 * fj
    f1 = p3 * r11 * ci1 + p4 * fdi1
    f2 = p3 * r22 * ci2 + p4 * fdi2

    s_val = sig2 * (f1 * coord1 + f2 * coord2)

    return s_val


@njit
def three_body_se_perm(r11, r12, r13, r21, r22, r23, r31, r32, r33, c1, c2,
                       ci1, ci2, ei1, ei2, ej1, ej2, fi, fj, fdi, ls1, ls2,
                       sig2, coord1, coord2, fdi1, fdi2):
    kern = 0

    if (c1 == c2):
        if (ei1 == ej1) and (ei2 == ej2):
            kern += three_body_se_helper(ci1, ci2, r11, r22, r33, fi, fj, fdi,
                                         ls1, ls2, sig2, coord1, coord2,
                                         fdi1, fdi2)
        if (ei1 == ej2) and (ei2 == ej1):
            kern += three_body_se_helper(ci1, ci2, r12, r21, r33, fi, fj, fdi,
                                         ls1, ls2, sig2, coord1, coord2,
                                         fdi1, fdi2)
    if (c1 == ej1):
        if (ei1 == ej2) and (ei2 == c2):
            kern += three_body_se_helper(ci1, ci2, r13, r21, r32, fi, fj, fdi,
                                         ls1, ls2, sig2, coord1, coord2,
                                         fdi1, fdi2)
        if (ei1 == c2) and (ei2 == ej2):
            kern += three_body_se_helper(ci1, ci2, r11, r23, r32, fi, fj, fdi,
                                         ls1, ls2, sig2, coord1, coord2,
                                         fdi1, fdi2)
    if (c1 == ej2):
        if (ei1 == ej1) and (ei2 == c2):
            kern += three_body_se_helper(ci1, ci2, r13, r22, r31, fi, fj, fdi,
                                         ls1, ls2, sig2, coord1, coord2,
                                         fdi1, fdi2)
        if (ei1 == c2) and (ei2 == ej1):
            kern += three_body_se_helper(ci1, ci2, r12, r23, r31, fi, fj, fdi,
                                         ls1, ls2, sig2, coord1, coord2,
                                         fdi1, fdi2)

    return kern


@njit
def three_body_grad_perm(r11, r12, r13, r21, r22, r23, r31, r32, r33, c1, c2,
                         ci1, ci2, cj1, cj2, ei1, ei2, ej1, ej2, fi, fj, fdi,
                         fdj, ls1, ls2, ls3, ls4, ls5, ls6, sig2, sig3):

    kern = 0
    sig_derv = 0
    ls_derv = 0

    if (c1 == c2):
        if (ei1 == ej1) and (ei2 == ej2):
            kern_term, sig_term, ls_term = \
                three_body_grad_helper_1(ci1, ci2, cj1, cj2, r11, r22, r33, fi,
                                         fj, fdi, fdj, ls1, ls2, ls3, ls4, ls5,
                                         ls6, sig2, sig3)
            kern += kern_term
            sig_derv += sig_term
            ls_derv += ls_term

        if (ei1 == ej2) and (ei2 == ej1):
            kern_term, sig_term, ls_term = \
                three_body_grad_helper_1(ci1, ci2, cj2, cj1, r12, r21, r33, fi,
                                         fj, fdi, fdj, ls1, ls2, ls3, ls4, ls5,
                                         ls6, sig2, sig3)
            kern += kern_term
            sig_derv += sig_term
            ls_derv += ls_term

    if (c1 == ej1):
        if (ei1 == ej2) and (ei2 == c2):
            kern_term, sig_term, ls_term = \
                three_body_grad_helper_2(ci2, ci1, cj2, cj1, r21, r13, r32, fi,
                                         fj, fdi, fdj, ls1, ls2, ls3, ls4, ls5,
                                         ls6, sig2, sig3)
            kern += kern_term
            sig_derv += sig_term
            ls_derv += ls_term

        if (ei1 == c2) and (ei2 == ej2):
            kern_term, sig_term, ls_term = \
                three_body_grad_helper_2(ci1, ci2, cj2, cj1, r11, r23, r32, fi,
                                         fj, fdi, fdj, ls1, ls2, ls3, ls4, ls5,
                                         ls6, sig2, sig3)
            kern += kern_term
            sig_derv += sig_term
            ls_derv += ls_term

    if (c1 == ej2):
        if (ei1 == ej1) and (ei2 == c2):
            kern_term, sig_term, ls_term = \
                three_body_grad_helper_2(ci2, ci1, cj1, cj2, r22, r13, r31, fi,
                                         fj, fdi, fdj, ls1, ls2, ls3, ls4, ls5,
                                         ls6, sig2, sig3)
            kern += kern_term
            sig_derv += sig_term
            ls_derv += ls_term

        if (ei1 == c2) and (ei2 == ej1):
            kern_term, sig_term, ls_term = \
                three_body_grad_helper_2(ci1, ci2, cj1, cj2, r12, r23, r31, fi,
                                         fj, fdi, fdj, ls1, ls2, ls3, ls4, ls5,
                                         ls6, sig2, sig3)

            kern += kern_term
            sig_derv += sig_term
            ls_derv += ls_term

    return kern, sig_derv, ls_derv


# -----------------------------------------------------------------------------
#                        many body helper functions
# -----------------------------------------------------------------------------

@njit


def k_sq_exp_double_dev(q1, q2, sig, ls):
    """Second Gradient of generic squared exponential kernel on two many body functions

    Args:
        q1 (float): the many body descriptor of the first local environment
        q2 (float): the many body descriptor of the second local environment
        sig (float): amplitude hyperparameter
        ls2 (float): squared lenghtscale hyperparameter
    Return:
        float: the value of the double derivative of the squared exponential kernel
    """

    qdiffsq = (q1 - q2) * (q1 - q2)

    ls2 = ls * ls

    ker = exp(-qdiffsq / (2 * ls2))

    ret = sig * sig * ker / ls2 * (1 - qdiffsq / ls2)

    return ret

@njit


def k_sq_exp_dev(q1, q2, sig, ls):
    """First Gradient of generic squared exponential kernel on two many body functions

    Args:
        q1 (float): the many body descriptor of the first local environment
        q2 (float): the many body descriptor of the second local environment
        sig (float): amplitude hyperparameter
        ls2 (float): squared lenghtscale hyperparameter
    Return:
        float: the value of the derivative of the squared exponential kernel
    """

    qdiff = (q1 - q2)

    ls2 = ls * ls

    ker = exp(-qdiff * qdiff / (2 * ls2))

    ret = - sig * sig * ker / ls2 * qdiff

    return ret


@njit
def coordination_number(rij, cij, r_cut, cutoff_func):
    """Pairwise contribution to many-body descriptor based on number of
        atoms in the environment

    Args:
        rij (float): distance between atoms i and j
        cij (float): Component of versor of rij along given direction
        r_cut (float): cutoff hyperparameter
        cutoff_func (callable): cutoff function
    Return:
        float: the value of the pairwise many-body contribution
        float: the value of the derivative of the pairwise many-body
        contribution w.r.t. the central atom displacement
    """

    fij, fdij = cutoff_func(r_cut, rij, cij)

    return fij, fdij


@njit
def q_value(distances, r_cut, cutoff_func, q_func=coordination_number):
    """Compute value of many-body descriptor based on distances of atoms
    in the local many-body environment.

    Args:
        distances (np.ndarray): distances between atoms i and j
        r_cut (float): cutoff hyperparameter
        cutoff_func (callable): cutoff function
        q_func (callable): many-body pairwise descrptor function

    Return:
        float: the value of the many-body descriptor
    """
    q = 0

    for d in distances:
        q_, _ = q_func(d, 0, r_cut, cutoff_func)
        q += q_

    return q

@njit
def q_value_mc(distances, r_cut, ref_species, species, cutoff_func, q_func=coordination_number):
    """Compute value of many-body many components descriptor based
    on distances of atoms in the local many-body environment.

    Args:
        distances (np.ndarray): distances between atoms i and j
        r_cut (float): cutoff hyperparameter
        ref_species (int): species to consider to compute the contribution
        species (np.ndarray): atomic species of neighbours
        cutoff_func (callable): cutoff function
        q_func (callable): many-body pairwise descrptor function

    Return:
        float: the value of the many-body descriptor
    """

    q = 0.0

    for i in range(len(distances)):
        if species[i] == ref_species:
            q_, _ = q_func(distances[i], 0, r_cut, cutoff_func)
            q += q_

    return q


@njit
def q_value_mc(distances, r_cut, ref_species, species, cutoff_func, q_func=coordination_number):
    """Compute value of many-body many components descriptor based
    on distances of atoms in the local many-body environment.

    Args:
        distances (np.ndarray): distances between atoms i and j
        r_cut (float): cutoff hyperparameter
        ref_species (int): species to consider to compute the contribution
        species (np.ndarray): atomic species of neighbours
        cutoff_func (callable): cutoff function
        q_func (callable): many-body pairwise descrptor function

    Return:
        float: the value of the many-body descriptor
    """

    q = 0

    for i in range(len(distances)):
        if species[i] == ref_species:
            q_, _ = q_func(distances[i], 0, r_cut, cutoff_func)
            q += q_

    return q


@njit
def mb_grad_helper_ls_(qdiffsq, sig, ls):
    """
    Derivative of a many body force-force kernel w.r.t. ls
    """

    ls2 = ls * ls

    prefact = exp(-(qdiffsq / (2 * ls2))) * (sig * sig) / ls ** 5

    ret = - prefact * (qdiffsq ** 2 / ls2 - 5 * qdiffsq + 2 * ls2)

    return ret

@njit
def mb_grad_helper_ls(q1, q2, qi, qj, sig, ls):
    """
    Helper function for many body gradient collecting all the derivatives
    of the force-force many body kernel w.r.t. ls
    """

    q12diffsq = ((q1 - q2) * (q1 - q2))
    qijdiffsq = ((qi - qj) * (qi - qj))
    qi2diffsq = ((qi - q2) * (qi - q2))
    q1jdiffsq = ((q1 - qj) * (q1 - qj))

    dk12 = mb_grad_helper_ls_(q12diffsq, sig, ls)
    dkij = mb_grad_helper_ls_(qijdiffsq, sig, ls)
    dki2 = mb_grad_helper_ls_(qi2diffsq, sig, ls)
    dk1j = mb_grad_helper_ls_(q1jdiffsq, sig, ls)

    return dk12 + dkij + dki2 + dk1j
